@import url("./bulma.css");

textarea {
	font-family:
		Consolas,
		Monaco,
		Lucida Console,
		Liberation Mono,
		DejaVu Sans Mono,
		Bitstream Vera Sans Mono,
		Courier New,
		monospace;
}

#output pre {
	padding: 0;
}

/* Override system theme preference when data-theme is set */
:root[data-theme="light"] {
	color-scheme: light;
}

:root[data-theme="dark"] {
	color-scheme: dark;
}

.title a {
	color: inherit;
	text-decoration: none;
}

.title a:hover {
	text-decoration: underline;
}
